local io = io

local csv = require "csv"
local dir = require "pl.dir"
local libmagic = require "libmagic"
local plpath = require "pl.path"
local pp = require "pl.pretty"

local term = require "lsup.term"
local triple = require "lsup.triple"
local graph = require "lsup.graph"

local mc = require "pocket_archive.monocypher"
local model = require "pocket_archive.model"
local pkar = require "pocket_archive"


local logger = pkar.logger

-- "nil" table - for missing key fallback in chaining.
local NT = {}

local M = {}  -- Submission module

-- Adapted from lua-núcleo
local function escape_pattern(s)
    local matches = {
        ["^"] = "%^";
        ["$"] = "%$";
        ["("] = "%(";
        [")"] = "%)";
        ["%"] = "%%";
        ["."] = "%.";
        ["["] = "%[";
        ["]"] = "%]";
        ["*"] = "%*";
        ["+"] = "%+";
        ["-"] = "%-";
        ["?"] = "%?";
        ["\0"] = "%z";
    }

    return (s:gsub(".", matches))
end

-- Initialize libmagic database.
local magic = libmagic.open( libmagic.MIME_TYPE, libmagic.NO_CHECK_COMPRESS )
assert(magic:load())


-- For idgen(). Makes a 60-character pool with ~5.9 bits of entropy per char.
local chpool = {}
for i = 48, 57  do table.insert(chpool, i) end  -- 0-9
for i = 65, 90  do table.insert(chpool, i) end  -- A-Z
for i = 97, 122 do table.insert(chpool, i) end  -- a-z

--[[
Generate a random, reader-friendly ID.

A 16-character ID with the above defined #chpool of 60 smybols has an entropy
of 94.5 bits, which should be plenty for a small repository.
]]
M.idgen = function(len)
    local charlist = {}
    for i = 1, (len or pkar.config.id.len) do
        table.insert(charlist, string.char(chpool[math.random(1, #chpool)]))
    end

    return table.concat(charlist)
end


--[=[
M.generate_sip_v1 = function(path)
    --[[
    Version 1 CSV parsing.

    Each row has 3 cells: reference path (ref), metadata key (k), value (v).
    Each v cell contains one value.

    Properties for the same ref can be added in successive rows without
    repeating the ref cell.

    Multiple values can be added in successive rows without repeating the ref
    and v cells.
    --]]

    local sub_data = assert(csv.open(path))
    local md = {}
    local prev_ref, prev_k

    -- Collate metadata.
    local i = 1
    for row in sub_data:lines() do
        ref, k, v = table.unpack(row)
        -- nil-out empty cells (they come through as "")
        if ref == "" then ref = nil end
        if k == "" then k = nil end
        if v == "" then v = nil end

        print("Parsing row:", ref, k, v)
        -- v can be a legit false value.
        if ref and not k and v == nil then
            -- This can be a placeholder for ordering purposes.
            md[ref] = md_ref or {}
            goto continue
        elseif v == nil then
            goto continue
        else
            -- If ref or k are missing, reuse the previous one.
            if ref then prev_ref = ref
            else
                if not prev_ref then
                    -- If column 1 is empty, it must have been set in a
                    -- previous row.
                    error(string.format(
                        "Reference in column 1, row %d not found!", i), 2)
                end
                ref = prev_ref
            end

            if k then prev_k = k
            else
                if not prev_k then
                    -- If column 2 is empty, it must have been set in a
                    -- previous row.
                    error(string.format(
                        "Property key in column 2, row %d not found!", i), 2)
                end
                k = prev_k
            end
        end

        md[ref] = md[ref] or {id = M.idgen(), path = ref, _sort = i}
        md[ref][k] = md[ref][k] or {}
        if k == "type" then
            md[ref][k] = v
        else
            table.insert(md[ref][k], v)
        end

        ::continue::
        i = i + 1
    end

    -- Move md to an ordered list.
    mdlist = {root_path = path:match("(.*/)")}
    for _, v in pairs(md) do table.insert(mdlist, v) end
    table.sort(mdlist, function (a, b) return (a._sort < b._sort) end)

    -- Infer structure from paths and row ordering.
    for i, v in ipairs(mdlist) do
        for j = i + 1, #mdlist do
            --print(string.format("comparing %s : %s", v["pas:sourcePath"], mdlist[j]["pas:sourcePath"]))
            if not v["next"] and
                    mdlist[j]["pas:sourcePath"]:match("(.*/)") == v["pas:sourcePath"]:match("(.*/)") then
                --print("next match.")
                v["next"] = mdlist[j]["pas:sourcePath"]
            end
            if not v.first and
                    mdlist[j]["pas:sourcePath"]:match("^" .. escape_pattern(v["pas:sourcePath"])) then
                --print("First child match.")
                v.first = mdlist[j]["pas:sourcePath"]
            end
        end
        v._sort = nil
    end

    return mdlist
end
--]=]


M.generate_sip_v2 = function(path)
    local sub_data = assert(csv.open(path, {header = true}))
    local sip = {root_path = path:match("(.*/)")}

    local prev_path

    local i = 1
    for row in sub_data:lines() do
        print("Processing row: " .. i)
        print("Row path: " .. row["pas:sourcePath"])
        if row["pas:sourcePath"] ~= "" then
            prev_path = row["pas:sourcePath"]
            -- New row.
            local id
            if row.id then
                id = "par:" .. row.id
                row.id = nil
            else id = "par:" .. M.idgen() end
            sip[i] = {id = id}
            for k, v in pairs(row) do
                if v == "" then goto cont1 end  -- skip empty strings.
                if pkar.config.md.single_values[k] then sip[i][k] = v
                else sip[i][k] = {[v] = true} end  -- Multi-values are a set.
                ::cont1::
            end
            i = i + 1
        else
            -- Continuation of values from a previous row.
            if i == 1 then
                error("First row MUST have a path value.", 2)
            elseif not prev_path then
                error(("No path information at row %d"):format(i), 2)
            else
                row["pas:sourcePath"] = prev_path
                for k, v in pairs(row) do
                    if v == "" then goto cont2 end  -- skip empty strings.
                    if pkar.config.md.single_values[k] then
                        -- It doesn't make much sense to overwrite, maybe throw an error?
                        sip[i - 1][k] = v
                    else
                        print("Value: " .. v)
                        print("Inserting at row " .. i -1)
                        sip[i - 1][k][v] = true
                    end
                    ::cont2::
                end
            end
        end
    end
    -- Infer structure from paths and row ordering.
    for i, v in ipairs(sip) do
        for j = i + 1, #sip do
            --print(string.format("comparing %s : %s", v["pas:sourcePath"], sip[j]["pas:sourcePath"]))
            if not v["pas:next"] and
                    sip[j]["pas:sourcePath"]:match("(.*/)") ==
                            v["pas:sourcePath"]:match("(.*/)") then
                --print("next match.")
                v["pas:next"] = sip[j].id
            end
            if not v["pas:first"] and
                    sip[j]["pas:sourcePath"]:match("^" .. escape_pattern(v["pas:sourcePath"])) then
                --print("First child match.")
                v["pas:first"] = sip[j].id
            end
        end
        v._sort = nil
    end
    --require "debugger"()

    return sip
end


M.validate = function(sip)
    -- TODO
end


M.update_rsrc_md = function(rsrc)
    -- TODO use a transaction when lsup_lua supports it.
    print("Updating resource md: ")
    pp.dump(rsrc)
    rmod = model.parse_model(rsrc["pas:contentType"])
    triples = {}

    gr = graph.new(pkar.store, term.DEFAULT_CTX)
    local s = term.new_iriref_ns(rsrc.id)
    for k, v in pairs(rsrc) do
        -- id is the subject, it won't be an attribute.
        if k == "id" then goto skip end

        print("Adding attribute:", k, v)
        local p = term.new_iriref_ns(k)
        local o
        if type(v) == "table" then
            for vv, _ in pairs(v) do
                if ((rmod.properties or NT)[k] or NT).type == "resource" then
                    o = term.new_iriref_ns(vv)
                else o = term.new_lit(vv) end
                table.insert(triples, triple.new(s, p, o))
            end
        else
            if ((rmod.properties or NT)[k] or NT).type == "resource" then
                o = term.new_iriref_ns(v)
            else o = term.new_lit(v) end
            table.insert(triples, triple.new(s, p, o))
        end
        ::skip::
    end
    for i, m in ipairs(rmod.lineage) do
        table.insert(
            triples, triple.new(s, pkar.RDF_TYPE, term.new_iriref_ns(m)))
    end

    -- This is a full replacement.
    --require "debugger"()
    logger:info("Removing triples.")
    gr:remove(s)
    logger:info("Adding triples.")
    -- TODO implement lsup_lua fn to add a single triple and add triples in
    -- the previous loop.
    gr:add(triples)
    logger:debug("Graph: ", tostring(gr))
    for trp in gr:lookup() do logger:debug(tostring(trp)) end
end


M.deposit = function(sip)
    for i, rsrc in ipairs(sip) do
        -- TODO Wrap this chunk into a txn. Each row is atomic.
        print(("Processing resource #%d of %d: %s"):format(i, #sip, rsrc.id))

        local in_path = sip.root_path .. rsrc["pas:sourcePath"]
        local fext = plpath.extension(in_path)
        -- If it's a directory, skip file processing.
        if not plpath.isfile(in_path) then goto continue end

        do
            local tmp_dir = pkar.config.fs.ores_path .. "tmp/"
            local file_ext
            _, file_ext = plpath.splitext(in_path)
            local tmp_path = tmp_dir .. rsrc.id .. file_ext
            dir.makepath(tmp_dir)

            local ifh = assert(io.open(in_path, "r"))

            rsrc["dc:format"] = {[magic:filehandle(ifh)] = true}
            local hash_it = mc.new_blake2b()
            local fsize = 0
            print(("Hashing %s"):format(in_path))
            local ofh = assert(io.open(tmp_path, "w"))
            while true do
                chunk = ifh:read(pkar.config.fs.stream_chunk_size)
                if not chunk then break end
                hash_it:update(chunk)
                ofh:write(chunk)
                fsize = fsize + #chunk
            end
            local checksum = hash_it:final(true)
            rsrc["premis:hasMessageDigest"] = {["blake2:" .. checksum] = true}
            rsrc["dc:extent"] = fsize

            ofh:close()
            ifh:close()

            local out_dir = ("%s%s/%s/"):format(
                    pkar.config.fs.ores_path,
                    checksum:sub(1, 4),
                    checksum:sub(5, 9))
            local out_path = out_dir .. checksum:sub(1,32) .. fext
            dir.makepath(out_dir)
            print(("Moving file %s to %s"):format(tmp_path, out_path))
            dir.movefile(tmp_path, out_path)
            rsrc["pas:path"] = out_path
        end

        ::continue::

        tstamp = os.date("!%Y-%m-%dT%TZ")
        rsrc["dc:created"] = tstamp
        rsrc["dc:modified"] = tstamp
        M.update_rsrc_md(rsrc)
    end
end


return M
